from dataclasses import dataclass
import datetime
import uuid
from typing import Optional


@dataclass
class Event:
    uuid: str = str(uuid.uuid4())
    timestamp: int = datetime.datetime.utcnow().timestamp()
    name: Optional[str] = None



from __future__ import annotations

from typing import Callable, Union, Any, Optional

from .event import Event
from .registry import Registry


class Mediator:
    def __init__(self,
                 registry: Registry = Registry()):
        self._registry = registry

    @staticmethod
    def _event_name(event: Union[Event, type]) -> str:
        if isinstance(event, type):
            return event.__name__
        elif isinstance(event, str):
            return event
        else:
            if event.name:
                return event.name
            else:
                return type(event).__name__

    def pub(self, event: Event) -> None:
        event_name = self._event_name(event)

        handlers = self._registry.get(event_name)

        for handler in handlers:
            handler(event)

    def add(self, event: Union[type, str], handler: Callable[[Any], None], priority: Optional[int] = None) -> Mediator:
        event_name = self._event_name(event)

        self._registry.add(event_name, handler, priority)

        return self



from abc import ABC
from typing import Optional

from .event import Event
from .mediator import Mediator


class Publisher(ABC):
    def __init__(self,
                 mediator: Optional[Mediator] = None):
        self._mediator = mediator

    def pub(self, event: Event):
        self._mediator.pub(event)



from typing import List, Callable, Dict, Optional

from .event import Event


class Registry:
    def __init__(self):
        self._handlers: Dict[str, List] = {}
        self._priorities: Dict[str, Dict[int, List[Callable[[Event], None]]]] = {}

    def add(self, event_name: str, handler: Callable[[Event], None], priority: Optional[int] = None) -> None:
        if event_name not in self._handlers:
            self._handlers[event_name] = []
            self._priorities[event_name] = []

        if not priority:
            if len(self._priorities[event_name]) > 0:
                priority = sorted(list(self._priorities[event_name].keys()))[-1] + 1
            else:
                priority = 1

        if priority not in self._priorities[event_name]:
            self._priorities[event_name][priority] = []

        self._priorities[event_name][priority].append(handler)

        priorities = sorted(list(self._priorities[event_name].keys()), reverse=True)

        self._handlers[event_name] = []

        for priority in priorities:
            handlers = self._priorities[event_name][priority]

            for handler in handlers:
                self._handlers[event_name].append(handler)

    def get(self, event: str) -> List[Callable[[Event], None]]:
        return self._handlers.get(event, [])



from __future__ import annotations

from typing import Optional, Callable, Any, Union

from .mediator import Mediator
from .publisher import Publisher


class Module(Publisher):
    def __init__(self, mediator: Optional[Mediator] = None):
        super().__init__(mediator)
        self._lazy_subscriptions = []

    def sub(self, event: Union[type, str], handler: Callable[[Any], None], priority: Optional[int] = None):
        if self._mediator:
            self._mediator.add(event, handler, priority)
        else:
            self._lazy_subscriptions.append((event, handler, priority))

        return self

    def init(self, mediator: Mediator):
        if not self._mediator:
            self._mediator = mediator
            if self._lazy_subscriptions:
                subscriptions = self._lazy_subscriptions
                self._lazy_subscriptions = []

                for event, handler, priority in subscriptions:
                    self.sub(event, handler, priority)

    def add(self, module: Module) -> Module:
        module.init(self._mediator)

        return self



# -*- coding: utf-8 -*-
import io

from setuptools import setup, find_packages

with io.open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

version = "0.0.5"

setup(
    name="decouple",
    version=version,
    description="Decoupling logic",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Andrey Avdeev",
    author_email="seorazer@gmail.com",
    license="Apache 2.0",
    packages=["decouple"],
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=[],
    keywords="decouple mediator decoupling",
    url="https://github.com/andrey-avdeev/decouple",
)



from .event import Event
from .mediator import Mediator
from .module import Module
from .publisher import Publisher
from .registry import Registry

__all__ = ["Event", "Mediator", "Module", "Publisher", "Registry"]
